#include <stdlib.h>
#include <stdio.h>
#include "arith.h"
#include "ppm.h"

/* pour le calcul ventuel de quelques statistiques */
/* #define STAT */

/* mise en mode debogage */
/* #define DEBUG */

/* nombre de noeuds maximum: peut tre modifi, mais doit tre tel que:
 * HASH_SIZE+NODE_NBMAX<65530
 */
#define NODE_NBMAX 40000

/*
 * taille des buffers d'entre/sortie
 */
#define BUFSIZE 1024

/*
 * Taille de la table de hachage
 */
#define HASH_SIZE 16384

/* ordre maximale de prdiction utilis (Phi) */
#define ORDER_MAX 4

/* frquence maximale d'un symbole */
#define RENORM_FREQSYM 250

/* frquence maximale pour la somme des frquences des symboles d'un contexte */ 
#define RENORM_FREQTOT 15000

/* nombre de couples symbole,frquence dans les structures de noeuds
 * ajust pour faire au total 16 octets */
#define NODE_SFNB 7
#define HDR_SFNB 2

/* nombre de symboles  coder sans compter les symboles spciaux */
#define SYM_NB 256

/* nombre de symboles spciaux: pour extension si besoin de signalisation */ 
#define SYM_SPECIAL_NB 1        

/* code associ au symbole ESCAPE (jamais cod de faon explicite) */
#define SYM_ESCAPE 256

/* code de fin de fichier */
#define SYM_EOF 256

/* valeur NULL pour les pointeurs stocks dans des USHORT */
#define NIL 0xFFFF

/* codage de NIL utilis pour retrouver l'adresse dans la table de hachage
 * d'un contexte en regardant seulement CTXHDR.hash_next */
#define HASH_ADDRESS (65530-HASH_SIZE)


/* stockage d'un couple symbole, frquence */
typedef struct {
   UCHAR sym;   /* numro du symbole */
   UCHAR freq;  /* frquence du symbole */
} SYMFREQ;


/* header pour grer un contexte */
typedef struct {
   USHORT ctx_next;        /* contexte suivant (moins utilis) */
	 USHORT ctx_prev;        /* contexte prcdent (plus utilis) */
   USHORT hash_next;       /* contexte suivant dans une entre de la table
														* de hachage */
   UCHAR order;            /* ordre du contexte */
   UCHAR sym[ORDER_MAX];   /* symboles constituant le contexte */
   UCHAR sf_max;           /* nombre de symboles-1 dans la liste L */ 
   union  {
      SYMFREQ sf[HDR_SFNB];   /* s'il y a moins de HDR_SFNB symboles dans
                               * le contexte, on les stocke directement ici
                               */
      struct {
				 USHORT freq_tot;     /* sinon on stocke la frquence totale (c) */ 
				 USHORT sf_next;      /* et un pointeur sur le premier noeud 
															 * constituant la liste L des symboles associs
															 * au contexte */
      } l;
   } sf;
} CTXHDR;

/* structure pour grer NODE_SFNB lments de la liste des symboles associs
 *  un contexte */
typedef struct {
	 SYMFREQ sf[NODE_SFNB];  /* les couples symbole, frquence */
	 USHORT sf_next;         /* pointeur sur l'ventuel bloc suivant */
} CTXSYMFREQ;

/*
 * structure de base pour la gestion de la mmoire: le noeud
 */

typedef union _NODE  {
   CTXHDR hdr;             /* si le noeud contient un header de contexte */
	 CTXSYMFREQ sf;          /* si le noeud contient une partie de la liste L */
   USHORT free_next;       /* si le noeud est vide: pointeur sur un noeud vide
														* suivant */
} NODE;

/* gestion des noeuds */
NODE *node_heap;           /* heap contenant tous les noeuds */           
UINT node_free_nb;         /* nombre de noeuds vides */
UINT node_free_first;      /* premier noeud de la liste des noeuds vides */
UINT node_free_last;       /* dernier noeud de la liste des noeuds vides */

/* gestion de la liste des contextes les plus utiliss */ 
USHORT *hash_table;        /* table de hachage pour rechercher un contexte */
UINT ctx_first;            /* premier contexte: le plus utilis */
UINT ctx_last;             /* dernier contexte: le moins utilis */
UINT ctx_nb;               /* nombre de contextes */
 
/* donnes caractrisant le contexte courant */
UCHAR sym_context[ORDER_MAX+1];    /* symboles prcdants le symbole en cours
																		* de codage */
int sym_hash[ORDER_MAX+1];         /* index dans la table de hachage 
																		* correspondants aux diffrentes longueurs
																		* de contextes
																		*/

/* systme d'exclusion des caractres */
UCHAR sym_excl[SYM_NB];            /* tableau pour l'exclusion */
UCHAR sym_excl_code;               /* code courant pour l'exclusion */


/* dclaration des fonctions de base */

/* noeuds */
static inline void Node_Free(UINT p);
static inline UINT Node_Alloc(void);

/* contextes */
static void Context_DeleteLast(void);
static void Context_MoveFirst(UINT c);
static void Context_New(int sym,int order);
static void Context_NewSym(int sym,UINT c);
static inline UINT Context_Search(int order);
static void Context_Renorm(UINT ctx);
static void Hash_Update(int sym);

/* init & end */
int PPM_Init(void);
void PPM_End(void);

/* codage */
void Encode_NewSym(int sym);
int Encode_NoExclude(int sym,UINT ctx);
int Decode_Exclude(UINT ctx);
int PPM_Decode(void);

/* dcodage */
int Decode_NewSym(void);
int Decode_NoExclude(UINT ctx);
int Decode_Exclude(UINT ctx);
int PPM_Decode(void);


#ifdef STAT
void PrintStat(void);
#endif


/****************************************************************************
 * Gestion des noeuds
 ****************************************************************************/

/* 
 * Dsallocation du noeud p
 */

static inline void Node_Free(UINT p) {
   node_heap[node_free_last].free_next=p;
   node_heap[p].free_next=NIL;
   node_free_last=p;
   node_free_nb++;
}

/*
 * Allocation d'un noeud
 * s'il ne reste plus de place, on dsalloue le contexte le moins utilis.
 */

static inline UINT Node_Alloc(void)  {
   UINT p;
   if (node_free_nb<=2) Context_DeleteLast();
   p=node_free_first;
   node_free_first=node_heap[node_free_first].free_next;
   node_free_nb--;
#ifdef DEBUG
   printf("Node_Alloc: p=%d\n",p);
#endif
   return p;
}

/****************************************************************************
 * Gestion des contextes 
 ****************************************************************************/


/* 
 * Mise au dbut de la liste des contextes du contexte c 
 */
static void Context_MoveFirst(UINT c) {
   NODE *ctx;
   
   if (c!=ctx_first)  {
      ctx=&node_heap[c];
      /* suppression du contexte dans la liste */
      if (c==ctx_last)  {
				 ctx_last=ctx->hdr.ctx_prev;
      } else  {
				 node_heap[ctx->hdr.ctx_prev].hdr.ctx_next=ctx->hdr.ctx_next;
				 node_heap[ctx->hdr.ctx_next].hdr.ctx_prev=ctx->hdr.ctx_prev;
      }
      /* insertion au dbut de la liste */
      node_heap[ctx_first].hdr.ctx_prev=c;
      ctx->hdr.ctx_next=ctx_first;
      ctx_first=c;
   }
}

/*
 * Destruction du contexte le moins utilis (ctx_last)
 */
static void Context_DeleteLast(void) {
   NODE *n;
   UINT h,h_next,node,node_next;
   USHORT *p;
   
   n=&node_heap[ctx_last];
   
   /* libration dans la table de hachage. Comme on ne dispose pas de
    * pointeur hash_prev dans les contextes, il faut parcourir toute
    * la liste. Heureusement, celle-ci est de longueur faible en moyenne
    */
   h_next=n->hdr.hash_next;
   h=h_next;
   while (h<HASH_ADDRESS) h=node_heap[h].hdr.hash_next;
   p=&hash_table[h-HASH_ADDRESS];
   while (*p!=ctx_last) p=&node_heap[*p].hdr.hash_next;
   *p=h_next;
   
   /* libration des noeuds & modification de ctx_last */
   
   if (n->hdr.sf_max>=2)  {
      node=n->hdr.sf.l.sf_next;
      while (1) {
				 node_next=node_heap[node].sf.sf_next;
				 Node_Free(node);
				 if (node_next==NIL) break;
				 node=node_next;
      }
   } 

   node=ctx_last;
   ctx_last=n->hdr.ctx_prev;
   Node_Free(node);
   ctx_nb--;
}

/* 
 * Cration d'un nouveau contexte avec un seul symbole sym de frquence 1
 * Utilisation implicite de sym_context et sym_hash.
 * Libration de mmoire si ncessaire, et mise en premier dans la liste
 */
static void Context_New(int sym,int order) {
   NODE *ctx;
   UINT i,c;

#ifdef DEBUG   
   printf("Context_New: sym=%d o=%d\n",sym,order);
#endif

   c=Node_Alloc();
   ctx=&node_heap[c];
   
   /* mise du contexte en tte de la liste */
   ctx->hdr.ctx_next=ctx_first;
   node_heap[ctx_first].hdr.ctx_prev=c;
   ctx_first=c;
   ctx_nb++;
   
   /* insertion dans la table de hachage */
   ctx->hdr.hash_next=hash_table[sym_hash[order]];
   hash_table[sym_hash[order]]=ctx_first;

   /* initialisation du contexte */
   ctx->hdr.order=order;
   for(i=0;i<order;i++) ctx->hdr.sym[i]=sym_context[i+1];

   ctx->hdr.sf_max=0;
   ctx->hdr.sf.sf[0].sym=sym;
   ctx->hdr.sf.sf[0].freq=1;
#ifdef DEBUG
   Context_Print(ctx_first);
#endif
}

/*
 * Ajout d'un nouveau symbole au contexte c
 */

static void Context_NewSym(int sym,UINT c) {
   NODE *n,*m;
   UINT p,sf_max;

#ifdef DEBUG   
   printf("Context_NewSym: sym=%d c=%d\n",sym,c);
   Context_Print(c);
#endif

   n=&node_heap[c];
   sf_max=n->hdr.sf_max;
   n->hdr.sf_max++;
   if (sf_max==0)  {
      n->hdr.sf.sf[1].sym=sym;
      n->hdr.sf.sf[1].freq=1;
   } else if (sf_max==1)  {
      p=Node_Alloc();
      m=&node_heap[p];
      m->sf.sf[0]=n->hdr.sf.sf[0];
      m->sf.sf[1]=n->hdr.sf.sf[1];
      m->sf.sf[2].sym=sym;
      m->sf.sf[2].freq=1;
      m->sf.sf_next=NIL;
      n->hdr.sf.l.sf_next=p;
      n->hdr.sf.l.freq_tot=((UINT)m->sf.sf[0].freq+(UINT)m->sf.sf[1].freq+1);
   } else  {
      n->hdr.sf.l.freq_tot++;
      m=&node_heap[n->hdr.sf.l.sf_next];
      while (sf_max>=NODE_SFNB)  {
				 sf_max-=NODE_SFNB;
				 m=&node_heap[m->sf.sf_next];
      }
      sf_max++;
      if (sf_max==NODE_SFNB)  {
				 p=Node_Alloc();
				 m->sf.sf_next=p;
				 m=&node_heap[p];
				 m->sf.sf_next=NIL;
				 sf_max=0;
      }
      m->sf.sf[sf_max].sym=sym;
      m->sf.sf[sf_max].freq=1;
   }
#ifdef DEBUG
   Context_Print(c);
#endif
}


#ifdef STAT
int hash_nb=1;
int hash_cnt=0;
#endif 

/*
 * Recherche d'un contexte, utilisation de faon implicite de sym_context
 * et de sym_hash.
 * C'est une procdure trs critique qui doit tre particulirement optimise
 */

static inline UINT Context_Search(int order)  {
   UCHAR *sym;
   UINT i,p;
   NODE *n;
#ifdef DEBUG   
   printf("Context_Search: o=%d\n",order);
#endif
   
   p=hash_table[sym_hash[order]];
	 sym=&sym_context[1];
#ifdef STAT
	 hash_nb++;
#endif
	 while (p<HASH_ADDRESS)  {
#ifdef STAT
			hash_cnt++;
#endif
			n=&node_heap[p];
			if (n->hdr.order==order)  {
				 if (order==0) return p;
				 i=0;
				 while (sym[i]==n->hdr.sym[i]) {
						i++;
						if (i==order) return p;
				 }
			}
			p=n->hdr.hash_next;
	 }
	 return HASH_ADDRESS;
}

/*
 * Cette macro est HORRIBLE mais permet de simplifier beaucoup le parcours
 * des listes de couples symbole,frquence tout en ayant un code rapide.
 * Une alternative lgante mais lente aurait t de passer une fonction
 * en paramtre contenant le code  excuter
 */

#define SF_Read(n,p,code_to_execute) \
{\
	 UINT nb,i;\
	 nb=(UINT)n->hdr.sf_max+1;\
   if (nb<=HDR_SFNB)  {\
      p=&n->hdr.sf.sf[0];\
   } else {\
      p=&node_heap[n->hdr.sf.l.sf_next].sf.sf[0];\
      while (nb>NODE_SFNB) {\
				 for(i=0;i<NODE_SFNB;i++)  {\
						code_to_execute;\
						p++;\
				 }\
				 p=&node_heap[ *((USHORT *)p) ].sf.sf[0];\
				 nb-=NODE_SFNB;\
      }\
   }\
   for(i=0;i<nb;i++)  {\
			code_to_execute;\
      p++;\
   }\
}


/* 
 * Renormalisation d'un contexte, ie, division de toutes les frquences 
 * par 2 et limination des symboles de frquence nulle  
 * Note: le contexte obtenu n'est jamais vide. 
 * Une amlioration prvue mais non implmente serait de trier le contexte
 * dans l'ordre des frquences dcroissantes pour acclrer la recherche.
 * Les gains en vitesse seraient de toute faon assez faibles car les
 * contextes sont de toute faon  peu prs tris vu leur mthode de 
 * construction: les caractres sont ajouts  la fin de la liste 
 */
static void Context_Renorm(UINT ctx) {
   NODE *n,*m;
   UINT a,b,c,i,freq_tot,sf_nb;
   SYMFREQ s,*p,tab_sf[SYM_NB];
	 
#ifdef DEBUG   
   printf("Context_Renorm: c=%d\n",ctx);
   Context_Print(ctx);
#endif 
   
   n=&node_heap[ctx];
	 freq_tot=0;
	 sf_nb=0;
	 
   SF_Read(n,p, {
			s=*p;
			s.freq=s.freq/2;
			if (s.freq!=0) {
				 freq_tot+=s.freq;
				 tab_sf[sf_nb]=s;
				 sf_nb++;
			}
	 } );
	 
	 
	 /* libration des noeuds utiliss pour stocker les symboles */
	 if (n->hdr.sf_max>=HDR_SFNB) {
			a=n->hdr.sf.l.sf_next;
			do {
				 b=node_heap[a].sf.sf_next;
				 Node_Free(a);
				 a=b;
			} while (a!=NIL);
	 }
				 
	 /* reconstruction de la liste des "sf_nb" symboles d'apres le tableau
    * "tab_sf"
    */ 

	 n->hdr.sf_max=sf_nb-1;
	 if (sf_nb<=HDR_SFNB)  {
			for(i=0;i<sf_nb;i++) n->hdr.sf.sf[i]=tab_sf[i];
	 } else  {
			a=Node_Alloc();
			n->hdr.sf.l.sf_next=a;
			n->hdr.sf.l.freq_tot=freq_tot;
			m=&node_heap[a];
			i=0;
			c=0;
			while (1) {
				 m->sf.sf[c]=tab_sf[i];
				 i++;
				 if (i==sf_nb) break;
				 c++;
				 if (c==NODE_SFNB)  {
						c=0;
						a=Node_Alloc();
						m->sf.sf_next=a;
						m=&node_heap[a];
				 }
			}
			m->sf.sf_next=NIL;
	 }

#ifdef DEBUG
   Context_Print(ctx);
#endif
}


/*
 * Mise  jour des index dans la table de hachage et des caractres du
 * contexte courant.
 * La fonction de hachage a t choisie de faon empirique en controlant
 * qu'elle donne en moyenne de bons rsultats.
 */
static void Hash_Update(int sym) {
   UINT i,k;
   
   for(i=ORDER_MAX;i>=2;i--) 
     sym_context[i]=sym_context[i-1];
   sym_context[1]=sym;
   
   for(i=ORDER_MAX;i>=2;i--) {  
      k=sym_hash[i-1];
      sym_hash[i]=( (k<<6)-k+sym ) & (HASH_SIZE-1);
   }
   sym_hash[1]=sym+1;
}


/****************************************************************************
 * Systme d'exclusion des symboles 
 ****************************************************************************/


/*
 * Remise  zro du tableau d'exclusion des symboles 
 */
static void Sym_ExcludeReset(void) {
   UINT i;
   
   sym_excl_code++;
   if (sym_excl_code==0)  {
      for(i=0;i<SYM_NB;i++) sym_excl[i]=0;
      sym_excl_code=1;
   }
}


/****************************************************************************
 * Initialisation et Libration mmoire 
 ****************************************************************************/

/*
 * Initialisation des structures de donnes du compresseur/dcompresseur
 * retourne 0 si tout va bien
 */
int PPM_Init(void) {
   UINT i;
   node_heap=malloc(NODE_NBMAX*sizeof(NODE));
	 hash_table=malloc(HASH_SIZE*sizeof(USHORT));
	 if (node_heap==NULL || hash_table==NULL)  {
			if (node_heap!=NULL) free(node_heap);
			if (hash_table!=NULL) free(hash_table);
			return 1;
	 }
	 
   /* noeuds: tous vides */
   for(i=0;i<=(NODE_NBMAX-2);i++) node_heap[i].free_next=i+1;
   node_heap[NODE_NBMAX-1].free_next=NIL;
   node_free_first=0;
   node_free_last=NODE_NBMAX-1;
   node_free_nb=NODE_NBMAX;
   
   /* contextes */
   for(i=0;i<HASH_SIZE;i++) hash_table[i]=HASH_ADDRESS+i;
   
   /* cette initialisation n'est pas sre mais simplifie beaucoup le code:
    * on suppose que le premier contexte sera allou dans le noeud 0
    */
   ctx_first=0;
   ctx_last=0;
   ctx_nb=0;
   
   /* contexte courant */
   for(i=0;i<=ORDER_MAX;i++) sym_context[i]=0;
   for(i=0;i<=ORDER_MAX;i++) sym_hash[i]=0;

   /* systme d'exclusion des caractres */
   sym_excl_code=0xFF;
	 
	 return 0;
}

/*
 * Fin de la compression/dcompression: on libre la mmoire
 */
void PPM_End(void) {
   free(hash_table);
   free(node_heap);
}


/****************************************************************************
 * Codage et Compression
 ****************************************************************************/

/*
 * Codage du symbole "sym" en considrant que tous les symboles sont de
 * frquence 1 sauf ceux exclus par sym_excl.
 */

void Encode_NewSym(int sym) {
   UINT freq,freq_tot,freq_cum,i;
   UCHAR code;
   
	 code=sym_excl_code;
	 if (sym>=SYM_NB)  {
			/* symbole spcial */
			freq_tot=0;
			for(i=0;i<SYM_NB;i++) if (sym_excl[i]!=code) freq_tot++;
			freq_cum=sym-(SYM_NB-freq_tot);
	 } else  {
			/* symbole "normal" */
			freq=0;
			for(i=0;i<sym;i++) if (sym_excl[i]!=code) freq++;
			freq_cum=freq;
			for(i=sym+1;i<SYM_NB;i++) if (sym_excl[i]!=code) freq++;
			freq_tot=freq+1;
	 }
   Arith_Encode(freq_cum,freq_cum+1,freq_tot+SYM_SPECIAL_NB);
}


/*
 * Codage du symbole "sym" en utilisant le contexte "ctx", et incrmentation
 * de sa frquence.
 * Retourne 0 si le codage a t effectu, et 1 si "sym" ne se trouve pas 
 * dans "ctx"
 */
int Encode_NoExclude(int sym,UINT ctx)  {
   UINT i,sf_nb,freq,freq_tot,freq_sym;
   SYMFREQ s,*p;
   UCHAR code;
   NODE *n;
   
#ifdef DEBUG   
   printf("Encode_NoExclude sym=%d c=%d\n",sym,ctx);
   Context_Print(ctx);
#endif

   n=&node_heap[ctx];
	 
	 /* Calcul de la somme des frquences des caractres */
	 if (n->hdr.sf_max<HDR_SFNB)  {
			freq_tot=0;
			for(i=0;i<=n->hdr.sf_max;i++) freq_tot+=n->hdr.sf.sf[i].freq;
	 } else  {
			freq_tot=n->hdr.sf.l.freq_tot;
	 }
	 
	 code=sym_excl_code;
	 
	 /* recherche du caractre en calculant la frquence */
   freq=0;
	 SF_Read(n,p, {
			s=*p;
			if (s.sym==sym) goto encode_sym;
			sym_excl[s.sym]=code;
			freq+=s.freq;
	 } );

	 /* symbole non trouv: on marque les caractres utiliss et 
		* on envoie le code "ESCAPE" */
   
	 code=sym_excl_code;
/*	 SF_Read(n,p, { sym_excl[p->sym]=code; }); */
   
	 sf_nb=(UINT) n->hdr.sf_max+1;
   Arith_Encode(freq_tot,freq_tot+sf_nb,freq_tot+sf_nb);
   return 1;

   encode_sym:   
   
   /* le symbole a t trouv: incrmentation frquence */
   freq_sym=s.freq;
   p->freq=freq_sym+1;
   if (n->hdr.sf_max>=HDR_SFNB) n->hdr.sf.l.freq_tot=freq_tot+1;

   sf_nb=(UINT) n->hdr.sf_max+1;
   Arith_Encode(freq,freq+freq_sym,freq_tot+sf_nb);
   
   /* test de la renormalisation */
   if (freq_sym==(RENORM_FREQSYM-1) || freq_tot>=RENORM_FREQTOT) { 
      Context_Renorm(ctx);
   }
   return 0;
}


/*
 * idem mais en excluant les symboles de sym_excl
 */

int Encode_Exclude(int sym,UINT ctx) {
   UINT freq_cum,freq_sym;
	 SYMFREQ *sf_p;
	 NODE *n;
	 register UCHAR code;
	 register SYMFREQ *p,s;
   register UINT freq_tot,sf_nb;
	 
#ifdef DEBUG   
   printf("Encode_NoExclude sym=%d c=%d\n",sym,ctx);
   Context_Print(ctx);
#endif

   n=&node_heap[ctx];
   code=sym_excl_code;

	 freq_tot=0;
	 sf_nb=0;
	 sf_p=NULL;
	 
	 SF_Read( n,p, {
			s=*p;
			if (sym_excl[s.sym]!=code) {
				 if (s.sym==sym) {
						freq_cum=freq_tot;
						sf_p=p;
				 }
				 sym_excl[s.sym]=code;
				 freq_tot+=s.freq;
				 sf_nb++;
			} 
	 } );
	 
	 if (sf_p==NULL)  {
			
			/* sortie d'un code ESCAPE */
			
			if (sf_nb!=0)  {
				 Arith_Encode(freq_tot,freq_tot+sf_nb,freq_tot+sf_nb);
			}
			return 1;
	 
	 } else  {
			
			/* incrmentation de la frquence */
			
			freq_sym=sf_p->freq;
			sf_p->freq=freq_sym+1;
			if (n->hdr.sf_max>=HDR_SFNB) n->hdr.sf.l.freq_tot++;
			
			/* codage du caractre */
			
			Arith_Encode(freq_cum,freq_cum+freq_sym,freq_tot+sf_nb);
   
			if (freq_sym==(RENORM_FREQSYM-1) || freq_tot>=RENORM_FREQTOT) { 
				 Context_Renorm(ctx);
			}
			
			return 0;
	 }
}


/*
 * Encodage d'un symbole: fonction de base 
 */
void PPM_Encode(int sym) {
   int i,e,order;
   UINT ctx,ctx_tab[ORDER_MAX+1],ctx_last;
   
   
   Sym_ExcludeReset();
   order=ORDER_MAX;
   ctx_last=NIL;
   while (1)  {
      ctx=Context_Search(order);
      ctx_tab[order]=ctx;
      if (ctx<HASH_ADDRESS)  {
				 Context_MoveFirst(ctx); 
				 if (ctx_last==NIL)
				 	e=Encode_NoExclude(sym,ctx);
				 else
				 	e=Encode_Exclude(sym,ctx); 
				 if (e==0) break;
				 ctx_last=ctx;
      } 
      order--;
      if (order==-1)  {
				 Encode_NewSym(sym);
				 if (sym>=SYM_NB) return;
				 break;
      }
   }

	 for(i=order+1;i<=ORDER_MAX;i++) { 
			if (ctx_tab[i]>=HASH_ADDRESS) 
			 Context_New(sym,i);
			else 
			 Context_NewSym(sym,ctx_tab[i]);
	 }
   
	 Hash_Update(sym); 
}


/*
 * Compression
 * Utilisation des fonction my_read_buf et my_write_buf pour les 
 * entres/sorties 
 */

int PPM_EncodeFile( UINT (*my_read_buf)(UCHAR *,UINT),
									 UINT (*my_write_buf)(UCHAR *,UINT) )  {
	 
   int i,len,err;
	 UCHAR *buf_in,*buf_out;
	 
	 buf_in=malloc(BUFSIZE);
	 buf_out=malloc(BUFSIZE);
	 
   err=0;
	 
	 if (buf_in==NULL || buf_out==NULL) {
			err=1;
	 } else  {
			Arith_EncodeInit(my_write_buf,buf_out,BUFSIZE);
			err=PPM_Init();
			if (err==0)  {
				 do  {
						len=my_read_buf(buf_in,BUFSIZE);
						for(i=0;i<len;i++) PPM_Encode(buf_in[i]);
				 } while (len==BUFSIZE);
				 
				 PPM_Encode(SYM_EOF);
				 
				 PPM_End();
			}
			Arith_EncodeEnd();
#ifdef STAT
			PPM_PrintStat();
#endif
	 }
	 
	 if (buf_in!=NULL) free(buf_in);
	 if (buf_out!=NULL) free(buf_out);

   return err;
}


/****************************************************************************
 * Dcodage et dcompression
 ****************************************************************************/

/*
 * Dcodage: cf Encode_NewSym
 */
int Decode_NewSym(void) {
	 UINT i,freq_tot,freq_cum,f;
	 UCHAR code;
	 
	 code=sym_excl_code;
	 freq_tot=0;
	 for(i=0;i<SYM_NB;i++) if (sym_excl[i]!=code) freq_tot++;
	 f=Arith_DecodeVal(freq_tot+SYM_SPECIAL_NB);
	 if (f>=freq_tot) {
			/* cas d'un symbole spcial */ 
			Arith_Decode(f,f+1,freq_tot+SYM_SPECIAL_NB);
			return SYM_NB+f-freq_tot;
	 } else  {
			i=0;
			freq_cum=0;
			while (1)  {
				 if (sym_excl[i]!=code)  {
						freq_cum++;
						if (freq_cum>f) break;
				 }
				 i++;
			}
			Arith_Decode(freq_cum-1,freq_cum,freq_tot+SYM_SPECIAL_NB);
			return i;
	 }
}

/*
 * Dcodage: cf Decode_NoExclude
 */
int Decode_NoExclude(UINT ctx) {
	 NODE *n;
	 UCHAR code;
	 UINT i,f,freq_tot,freq_cum,freq_sym,sf_nb;
	 SYMFREQ *p,s;
	 
	 
   n=&node_heap[ctx];
	 code=sym_excl_code;
	 
	 /* Calcul de la somme des frquences des caractres */
	 if (n->hdr.sf_max<HDR_SFNB)  {
			freq_tot=0;
			for(i=0;i<=n->hdr.sf_max;i++) freq_tot+=n->hdr.sf.sf[i].freq;
	 } else  {
			freq_tot=n->hdr.sf.l.freq_tot;
	 }
	 
	 /* dcodage */
	 sf_nb=(UINT) n->hdr.sf_max+1;
	 f=Arith_DecodeVal(freq_tot+sf_nb);
	 if (f>=freq_tot)  {
			/* gestion du code ESCAPE */
			
			/* marquage des caractres utiliss */
			SF_Read(n,p, { sym_excl[p->sym]=code; });
				 
			/* dcodage ESCAPE */
			Arith_Decode(freq_tot,freq_tot+sf_nb,freq_tot+sf_nb);
			return SYM_ESCAPE;
	 }
	 
	 /* recherche du caractre en calculant la frquence */
   freq_cum=0;
	 SF_Read(n,p, {
			s=*p;
			freq_cum+=s.freq;
			if (freq_cum>f) goto decode_sym;
	 } );
	 
	 decode_sym:

   freq_sym=s.freq;
   p->freq=freq_sym+1;
   if (n->hdr.sf_max>=HDR_SFNB) n->hdr.sf.l.freq_tot=freq_tot+1;

   Arith_Decode(freq_cum-freq_sym,freq_cum,freq_tot+sf_nb);
   
   /* test de la renormalisation */
   if (freq_sym==(RENORM_FREQSYM-1) || freq_tot>=RENORM_FREQTOT) { 
      Context_Renorm(ctx);
   }
   return s.sym;
}


/*
 * Dcodage: cf Encode_Exclude
 */

int Decode_Exclude(UINT ctx) {
	 UINT sf_nb,freq_sym,freq_cum,freq_tot,f;
	 NODE *n;
	 SYMFREQ s,*p;
	 UCHAR code;
	 
	 n=&node_heap[ctx];
   code=sym_excl_code;

	 freq_tot=0;
	 sf_nb=0;

	 SF_Read( n,p, {
			s=*p;
			if (sym_excl[s.sym]!=code) 
			{
				 freq_tot+=s.freq;
				 sf_nb++;
			}
	 } );
	 
	 
	 f=Arith_DecodeVal(freq_tot+sf_nb);
	 
	 if (f>=freq_tot) {
			
			/* ESCAPE */
			
			SF_Read(n,p, { sym_excl[p->sym]=code; } );
			
			Arith_Decode(freq_tot,freq_tot+sf_nb,freq_tot+sf_nb);
			
			return SYM_ESCAPE;
	 } else  {
			
			/* recherche du caractre */
			
			freq_cum=0;
			SF_Read(n,p, {
				 s=*p;
				 if (sym_excl[s.sym]!=code)  {
						freq_cum+=s.freq;
						if (freq_cum>f) goto decode_sym;
				 }
			} );
			
			decode_sym:
			
			/* incrmentation de la frquence */
			
			freq_sym=p->freq;
			p->freq=freq_sym+1;
			if (n->hdr.sf_max>=HDR_SFNB) n->hdr.sf.l.freq_tot++;
			
			/* dcodage du caractre */
			
			Arith_Decode(freq_cum-freq_sym,freq_cum,freq_tot+sf_nb);
   
			if (freq_sym==(RENORM_FREQSYM-1) || freq_tot>=RENORM_FREQTOT) { 
				 Context_Renorm(ctx);
			}
			
			return s.sym;
	 }
}


/*
 * Dcodage d'un symbole
 */
int PPM_Decode(void) {
   int i,order,sym;
   UINT ctx,ctx_tab[ORDER_MAX+1],ctx_last;
   
   
   /* recherche de l'ordre maximum */

   Sym_ExcludeReset();
   order=ORDER_MAX;
   ctx_last=NIL;
   while (1)  {
      ctx=Context_Search(order);
      ctx_tab[order]=ctx;
      if (ctx<HASH_ADDRESS)  {
				 Context_MoveFirst(ctx); 
				 if (ctx_last==NIL)
				 	sym=Decode_NoExclude(ctx);
				 else
				 	sym=Decode_Exclude(ctx);
				 if (sym!=SYM_ESCAPE) break;
				 ctx_last=ctx;
      } 
      order--;
      if (order==-1)  {
				 sym=Decode_NewSym();
				 if (sym>=SYM_NB) return sym;
				 break;
      }
   }

	 for(i=order+1;i<=ORDER_MAX;i++) { 
			if (ctx_tab[i]>=HASH_ADDRESS) 
			 Context_New(sym,i);
			else 
			 Context_NewSym(sym,ctx_tab[i]);
	 }
   
	 Hash_Update(sym);

	 return sym;
}
   

/*
 * Dcompression: idem
 */

int PPM_DecodeFile( UINT (*my_read_buf)(UCHAR *,UINT),
									 UINT (*my_write_buf)(UCHAR *,UINT) ) {

   int c,len,err;
	 UCHAR *buf_in,*buf_out;
	 
	 buf_in=malloc(BUFSIZE);
	 buf_out=malloc(BUFSIZE);
	 
   err=0;
	 
	 if (buf_in==NULL || buf_out==NULL) {
			err=1;
	 } else  {
			
			Arith_DecodeInit(my_read_buf,buf_in,BUFSIZE);
			err=PPM_Init();
			if (err==0)  {
				 len=0;
				 while (1) {
						c=PPM_Decode();
						if (c==SYM_EOF) break;
						buf_out[len++]=c;
						if (len==BUFSIZE)  {
							 my_write_buf(buf_out,len);
							 len=0;
						}
				 } 
				 if (len>0) my_write_buf(buf_out,len);
			
				 PPM_End();
			}
			Arith_DecodeEnd();
	 }
	 
	 if (buf_in!=NULL) free(buf_in);
	 if (buf_out!=NULL) free(buf_out);

   return err;
}
	 




#ifdef STAT

/****************************************************************************
 * Statistiques 
 ****************************************************************************/


void PrintStat(void)  {
	 fprintf(stderr,"free=%d ctx_nb=%d hash_moy=%0.2f\n",
					 node_free_nb,ctx_nb,
					 (float)hash_cnt/(float)hash_nb);
	 
}

/*
 * Impression d'un caractre
 */
void Sym_Print(int c) {
   if (c>=32 && c<=126) printf("%c",c); else printf("\\%2X",c);
}

/*
 * Impression couple SYMFREQ
 */

void SF_Print(SYMFREQ s) {
   Sym_Print(s.sym);
   printf(":%d ",s.freq);
}

/*
 * Impression du contenu d'un contexte
 * utilis pour les tests 
 */

void Context_Print(UINT c) {
   NODE *n;
   int i,sf_max,sf_nb,sf_freq;
   
   n=&node_heap[c];
   
   sf_max=n->hdr.sf_max;
   sf_nb=sf_max+1;
   if (sf_max>=2) sf_freq=n->hdr.sf.l.freq_tot;
   else  {
      sf_freq=0;
      for(i=0;i<=sf_max;i++) sf_freq+=n->hdr.sf.sf[i].freq;
   }

   printf("Ctx=%d: hash_n=%d ctx_p=%d ctx_n=%d o=%d sf_nb=%d sf_freq=%d\n",
	  c,n->hdr.hash_next,n->hdr.ctx_prev,n->hdr.ctx_next,
	  n->hdr.order,sf_nb,sf_freq);
   for(i=0;i<n->hdr.order;i++) Sym_Print(n->hdr.sym[i]);
   printf(": ");
   if (sf_max<=1) {
      for(i=0;i<=sf_max;i++) SF_Print(n->hdr.sf.sf[i]);
   } else  {
      n=&node_heap[n->hdr.sf.l.sf_next];
      i=0;
      while (1) {
	 SF_Print(n->sf.sf[i]);
	 if (sf_max==0) break;
	 i++;
	 sf_max--;
	 if (i==NODE_SFNB)  {
	    i=0;
	    n=&node_heap[n->sf.sf_next];
	 }
      }
   }
   printf("\n");
}


/*
 * Nombre total de contextes et nombre de contextes de longueur donnes.
 * Utilis pour les statistiques
 */

void Context_Statistic(void) {
   UINT i,p;
   int tab[SYM_NB+1],tot,cnt;
   
   for(i=0;i<=SYM_NB;i++) tab[i]=0;
   tot=0;
   
   p=ctx_first;
   do {
      cnt=node_heap[p].hdr.sf_max+1;
      tab[cnt]++;
      tot++;
      p=node_heap[p].hdr.ctx_next;
   } while (p!=ctx_last);
   
   
   printf("Context_Statistic: ");
   for(i=1;i<=SYM_NB;i++)  {
      printf("%d:%d (%0.2f%%),",i,tab[i],(float)tab[i]/(float)tot*100.0);
   }
   printf("\n");
}

#endif

